from tycho.config import Config
from tycho.factory import ComputeFactory
from tycho.factory import supported_backplanes
from tycho.model import System, ModifySystem

class Tycho: 
    """ An organizing abstraction for the Tycho system. 

        Tycho adds a layer of system architecture and policy support to 
        cloud native container orchestration platforms. It's true you can do just
        about anything with the Kubernetes API. Tycho let's teams design, decide,
        automate, test, and enforce what should be done.

    """
    
    def __init__(self,
                 backplane="kubernetes",
                 config="conf/tycho.yaml"):
        """ Construct a Tycho component. 
        
            :param backplane: The name of the compute back end. Analogous to a compiler's
                              code emitter, the backplane's concern is to project a system
                              abstract syntax tree (AST) into a compute fabric specific 
                              structure.
            :param config: A configuration file for the system.
        """
        self.backplane = backplane
        self.config = Config (config)
        self.compute = ComputeFactory.create_compute (self.config)
        
    def get_compute (self):
        """ Get the Tycho API for the compute fabric. 
    
            :returns: A compute fabric code emitter implementation specified to the constructor.
        """
        return self.compute

    def parse (self, request):
        """ Parse a request to construct an abstract syntax tree for a system.
        
            :param request: JSON object formatted to contain name, structure, env, and
                            service elements. Name is a string. Structue is the JSON
                            object resulting from loading a docker-compose.yaml. Env
                            is a JSON dictionary mapping environment variables to
                            values. These will be substituted into the specification.
                            Services is a JSON object representing which containers and
                            ports to expose, and other networking rules.
            :returns: `.System`
        """
        return System.parse (
            config=self.config,
            name=request['name'],
            principal=request.get('principal'),
            system=request['system'],
            service_account=request.get('serviceaccount', 'default'),
            env=request.get ('env', {}),
            services=request.get ('services', {}))

    def parse_modify(self, request):
        """ Parse a request into a class representation of metadata and specs of a system to be modified.

            :param request: JSON object formatted to contain guid, labels, resources.
                            GUID is a hexadecimal string of UUID representing a system.
                            Labels is a dictionary of label and label-name as key-value pairs.
                            Resources is a dictionary of cpu and memory keys, with corresponding values.
                            Can optionally pass a config.
            :returns: An instance of `tycho.model.ModifySystem`
        """
        return ModifySystem.parse_modify(
            config=self.config,
            guid=request.get("tycho-guid", None),
            labels=request.get("labels", {}),
            cpu=request.get("cpu", None),
            memory=request.get("memory", None))

    @staticmethod
    def is_valid_backplane (backplane):
        """ Determine if the argument is a valid backplane. """
        return ComputeFactory.is_valid_backplane (backplane)
    
    @staticmethod
    def supported_backplanes ():
        return list(supported_backplanes)



import datetime
import json
import logging
import netifaces
import os
import string
import traceback
import yaml
from jinja2 import Template

logger = logging.getLogger (__name__)

class TemplateUtils:
    """ Utilities for generating text. """

    def __init__(self, config):
        self.config = config

    def render (self, template, context):
        """Render a template object given a context. """
        result=None
        template_path = None
        
        """ First, allow the user to override the default templates with custom templates.
            Check for a template with this name in the user provided paths. """
        alternate_paths = self.config['tycho']['templates']['paths']
        for path in alternate_paths:
            if os.path.exists (path):
                template_path = os.path.join (path, template)
                if os.path.exists (template_path):
                    logger.debug (f"using user supplied template: {template_path}")
                else:
                    template_path = None
            else:
                logger.warning (f"template path {path} is configured but does not exist.")

        if not template_path:
            """ Still no template. Look for it in the default design. """
            template_path = os.path.join (os.path.dirname (__file__), "template", template)
            if not os.path.exists (template_path):
                template_path = None

        if not template_path:
            raise ValueError (
                f"No template {template} found in default location or in {alternate_paths}")
        
        logger.debug (f"applying template {template_path}")
        with open(template_path, "r") as stream:
            template_text = stream.read ()
            result = TemplateUtils.render_text (template_text, context)
        return result

    @staticmethod
    def render_text (template_text, context):
        """ Render the text of a template given a context. """
        #logger.debug (template_text)
        #logger.debug (context)
        template = Template (template_text)
        template.globals['now'] = datetime.datetime.utcnow
        text = template.render (**context)
        logger.debug (text)
        return yaml.load_all (text)

    @staticmethod
    def apply_environment (environment, text):
        """ Given an environment configuration consisting of lines of Bash style variable assignemnts,
        parse the variables and apply them to the given text."""
        resolved = text
        if environment:
            mapping = {
                line.split("=", maxsplit=1)[0] : line.split("=", maxsplit=1)[1]
                for line in environment.split ("\n") if '=' in line
            }
            resolved = string.Template(text).safe_substitute (**mapping)
            logger.debug (f"environment={json.dumps (mapping, indent=2)}")
            logger.debug (resolved)
        return resolved

    @staticmethod
    def trunc (a_string, max_len=80):
        return (a_string[:max_len] + '..') if len(a_string) > max_len else a_string

class NetworkUtils:
    @staticmethod
    def get_client_ip (request, debug=False):
        """ Get the IP address of the client. Account for requests from proxies. 
        In debug mode, ignore loopback and try to get an IP from a n interface."""
        ip_addr = request.remote_addr
        if request.headers.getlist("X-Forwarded-For"):
            ip_addr = request.headers.getlist("X-Forwarded-For")[0]
        if debug:
            interface = netifaces.ifaddresses ('en0')
            ip_addr = interface[2][0]['addr']
        logger.debug (f"(debug mode ip addr:)--> {ip_addr}")
        return ip_addr

class Resource:
    @staticmethod
    def get_resource_path(resource_name):
       # Given a string resolve it to a module relative file path unless it is already an absolute path.
        resource_path = resource_name
        if not resource_path.startswith (os.sep):
            resource_path = os.path.join (os.path.dirname (__file__), resource_path)
        return resource_path
    
    @staticmethod
    def load_json (path):
        result = None
        with open (path, 'r') as stream:
            result = json.loads (stream.read ())
        return result

    @staticmethod
    def load_yaml (path):
        result = None
        with open (path, 'r') as stream:
            result = yaml.safe_load (stream.read ())
        return result
    
    def get_resource_obj (resource_name, format=None):
        result = None
        if not format:
            if resource_name.endswith ('.yaml'):
                format = 'yaml'
            else:
                format = 'json'
        path = Resource.get_resource_path (resource_name)
        if os.path.exists (path):
            m = {
                'json' : Resource.load_json,
                'yaml' : Resource.load_yaml
            }
            if format in m:
                result = m[format](path)
        return result



import argparse
import json
import logging
import os
import sys
import threading
import traceback
import yaml

logger = logging.getLogger (__name__)

class Compute:
    """ Abstraction of a compute cluster. 

    We start with three primitives: start, status, and delete. Start takes a JSON object
    including (a) a docker-compose formatted description of a cloud native distributed system (b) the contents of an .env file accompanying a docker-compose.yaml including enviroment specific settings, and (c) additional metadata.
    
    """
    
    def start (self, system, namespace="default"):
        """ Given a system definition, start a distributed system.
        
            In docker port mapping pairs of the form <host_port>:<container_port>, we ignore 
            the host port. Tycho is designed to start many instances of an application and dynamic
            port allocation is assumed.

            Volume mounts of the form <host_path>:<container_path> will make a platform specific
            mapping of host_path. The general request format is::

                {
                   "name"   : <name of the system>,
                   "env"    : <text of .env environment variables>,
                   "system" : <JSON of a docker-compose yaml>
                }

            Responses contain status, message, and result elements.::

                {
                  "status": "success",
                  "result": {
                    "name": "nginx-7703f9cbf8f34caf8bc64e84384b7f1f",
                    "sid": "7703f9cbf8f34caf8bc64e84384b7f1f",
                    "containers": {
                      "nginx": {
                        "port": 30306
                      }
                    }
                  },
                  "message": "Started system nginx-7703f9cbf8f34caf8bc64e84384b7f1f"
                }

            :param system: docker-compose formatted specification of a distributed system.
            :type json: A JSON object structured as:
            :param namespace: Namespace. May not be supported by underlying compute fabric.
            :type namespace: string
            :return: Returns a JSON object including status, message, and result. Result is a dictionary containing details of the creatd object including name, system id (sid), and port mappings for each exposed service.

        """
        pass
    def delete (self, guid, namespace="default"):
        """ Delete a distributed system.
        
            Delete all generated artifacts in the underlying system.

            An example response::
         
                {
                  "status": "success",
                  "result": {},
                  "message": "Deleted system 82a9b5dc7d7c40c69ac05e3fb0f4df86"
                }

            :param guid: Globally unique identifier of the system, as returned by start.
            :type guid: string
            :param namespace: Namespace. May not be supported by underlying compute fabric.
            :type namespace: string
        """
        pass
    def status (self, guid=None, namespace="default"):
        """ Get status on running components of the system.

            Provided with a GUID, return status on matching components.

            Without a GUID, return data on all running components.

            An example response::

                {
                  "status": "success",
                  "result": [{
                    "name": "nginx-7703f9cbf8f34caf8bc64e84384b7f1f",
                    "sid": "7703f9cbf8f34caf8bc64e84384b7f1f",
                    "port": "30306"
                  }],
                  "message": "Get status for system None"
                }

            :param guid: GUID returned by start for a system.
            :param namespace: Namespace. May not be supported by underlying compute fabric.
        
        """
        pass



import argparse
import ipaddress
import json
import jsonschema
import logging
import netifaces
import os
import requests
import sys
import traceback
import yaml
from tycho.core import Tycho
from tycho.tycho_utils import NetworkUtils

"""
Provides actions for creating, monitoring, and deleting distributed systems of cloud native
containers running on abstracted compute fabrics. 
 
"""
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

""" Load the schema. """
schema_file_path = os.path.join (
    os.path.dirname(__file__),
    'api-schema.yaml')
template = None
with open(schema_file_path, 'r') as file_obj:
    template = yaml.load(file_obj)

backplane = None
_tycho = Tycho(backplane=backplane)


def tycho ():
    return _tycho


class TychoResource:
    """ Base class handler for Tycho requests. """
    def __init__(self):
        self.specs = {}
        
    """ Functionality common to Tycho services. """
    def validate(self, request, component):
        """ Validate a request against the schema. """
        if not self.specs:
            with open(schema_file_path, 'r') as file_obj:
                self.specs = yaml.load(file_obj)
        to_validate = self.specs["components"]["schemas"][component]
        try:
            logger.debug(f"--:Validating obj {request}")
            logger.debug(f"  schema: {json.dumps(to_validate, indent=2)}")
            jsonschema.validate(request, to_validate)
        except jsonschema.exceptions.ValidationError as error:
            logger.error(f"ERROR: {str(error)}")
            traceback.print_exc (error)

    def create_response(self, result=None, status='success', message='', exception=None):
        """ Create a response. Handle formatting and modifiation of status for exceptions. """
        if exception:
            traceback.print_exc()
            status = 'error'
            exc_type, exc_value, exc_traceback = sys.exc_info()
            message = f"{exception.args[0]} {''.join (exception.args[1])}" \
                if len(exception.args) == 2 else exception.args[0]
            result = {
                'error': message
            }
        return {
            'status': status,
            'result': result,
            'message': message
        }


class StartSystemResource(TychoResource):
    """ Parse, model, emit orchestrator artifacts and execute a system. """
    
    """ System initiation. """
    def post(self, request):
        response = {}
        try:
            logger.info(f"start-system: {json.dumps(request, indent=2)}")
            self.validate(request, component="System")
            system = tycho().parse(request)
            response = self.create_response(
                result=tycho().get_compute().start(system),
                message=f"Started system {system.name}")
        except Exception as e:
            response = self.create_response(
                exception=e,
                message=f"Failed to create system.")
        return response


class DeleteSystemResource(TychoResource):
    """ System termination. Given a GUID for a Tycho system, use Tycho core to eliminate all
    components comprising the running system."""
    def post(self, request):
        response = {}
        system_name = None
        try:
            logger.debug(f"delete-request: {json.dumps(request, indent=2)}")
            self.validate(request, component="DeleteRequest")
            system_name = request['name']
            response = self.create_response(
                result=tycho().get_compute().delete(system_name),
                message=f"Deleted system {system_name}")
        except Exception as e:
            response = self.create_response(
                exception=e,
                message=f"Failed to delete system {system_name}.")
        return response


class StatusSystemResource(TychoResource):
    """ Status executing systems. Given a GUID (or not) determine system status. """

    def post(self, request):
        response = {}
        try:
            logging.debug(f"list-request: {request}")
            self.validate(request, component="StatusRequest")
            system_name = request.get('name', None)
            system_username = request.get('username', None)
            response = self.create_response(
                result=tycho().get_compute().status(system_name, system_username),
                message=f"Get status for system {system_name}")
        except Exception as e:
            response = self.create_response(
                exception=e,
                message=f"Failed to get system status.")
        print(json.dumps(response, indent=2))
        return response


class ModifySystemResource(TychoResource):
    """ Modify a system given a name, labels, resources(cpu and memory) """

    def post(self, request):
        try:
            logging.debug(f"System specs to modify: {request}")
            system_modify = tycho().parse_modify(request)
            response = self.create_response(
                result=tycho().get_compute().modify(system_modify),
                message=f"Modified the system")
        except Exception as e:
            response = self.create_response(
                exception=e,
                message=f"Failed to modify system status.")
        return response



import argparse
import ipaddress
import json
import jsonschema
import logging
import netifaces
import os
import requests
import sys
import traceback
import yaml
from flasgger import Swagger
from flask import Flask, jsonify, g, Response, request, Request
from flask_restful import Api, Resource
from flask_cors import CORS
from tycho.core import Tycho
from tycho.tycho_utils import NetworkUtils

"""
Defines the Tycho API

Provides endpoints for creating, monitoring, and deleting distributed systems of cloud native
containers running on abstracted compute fabrics. 
 
"""
logger = logging.getLogger (__name__)

app = Flask(__name__)

""" Enable CORS. """
api = Api(app)
CORS(app)
debug=False

""" Load the schema. """
schema_file_path = os.path.join (
    os.path.dirname (__file__),
    'api-schema.yaml')
template = None
with open(schema_file_path, 'r') as file_obj:
    template = yaml.load(file_obj)

""" Describe the API. """
app.config['SWAGGER'] = {
    'title': 'Tycho Compute API',
    'description': 'An API, compiler, and executor for cloud native distributed systems.',
    'uiversion': 3
}

swagger = Swagger(app, template=template)

backplane = None
_tycho = Tycho(backplane=backplane)

def tycho ():
    return _tycho
#    if not hasattr(g, 'tycho'):
    if not 'tycho' in g: #hasattr(g, 'tycho'):
        logger.debug (f"-----------> {dir(g)}")
        logger.debug (f"--------------------------------> creating tycho object.")
        g.tycho = Tycho (backplane=backplane)
        logger.debug (f"-----------> {dir(g)}")
        logger.debug (f"--------------------------------> creating tycho object.")
    return g.tycho
    
class TychoResource(Resource):
    """ Base class handler for Tycho API requests. """
    def __init__(self):
        self.specs = {}
        
    """ Functionality common to Tycho services. """
    def validate (self, request, component):
        """ Validate a request against the schema. """
        if not self.specs:
            with open(schema_file_path, 'r') as file_obj:
                self.specs = yaml.load(file_obj)
        to_validate = self.specs["components"]["schemas"][component]
        try:
            app.logger.debug (f"--:Validating obj {json.dumps(request.json, indent=2)}")
            app.logger.debug (f"  schema: {json.dumps(to_validate, indent=2)}")            
            jsonschema.validate(request.json, to_validate)
        except jsonschema.exceptions.ValidationError as error:
            app.logger.error (f"ERROR: {str(error)}")
            traceback.print_exc (error)
            abort(Response(str(error), 400))

    def create_response (self, result=None, status='success', message='', exception=None):
        """ Create a response. Handle formatting and modifiation of status for exceptions. """
        if exception:
            traceback.print_exc ()
            status='error'
            exc_type, exc_value, exc_traceback = sys.exc_info()
            message = f"{exception.args[0]} {''.join (exception.args[1])}" \
                if len(exception.args) == 2 else exception.args[0]
            result = {
                'error' : message #str(exception) #repr(traceback.format_exception(exc_type, exc_value, exc_traceback))
            }
            print (json.dumps(result, indent=2))
        return {
            'status'  : status,
            'result'  : result,
            'message' : message
        }
            
class StartSystemResource(TychoResource):
    """ Parse, model, emit orchestrator artifacts and execute a system. """
    
    """ System initiation. """
    def post(self):
        """
        Start a system based on a specification on the compute fabric.
        
        The specification is a docker-compose yaml parsed into a JSON object.
        ---
        tag: start
        description: Start a system on the compute fabric.
        requestBody:
            description: System start message.
            required: true
            content:
                application/json:
                    schema:
                        $ref: '#/components/schemas/System'
        responses:
            '200':
                description: Success
                content:
                    text/plain:
                        schema:
                            type: string
                            example: "Successfully validated"
            '400':
                description: Malformed message
                content:
                    text/plain:
                        schema:
                            type: string

        """
        response = {}
        try:
            app.logger.info (f"start-system: {json.dumps(request.json, indent=2)}")
            self.validate (request, component="System")
            system = tycho().parse (request.json)
            response = self.create_response (
                result=tycho().get_compute().start (system),
                message=f"Started system {system.name}")
        except Exception as e:
            response = self.create_response (
                exception=e,
                message=f"Failed to create system.")
        return response
    
class DeleteSystemResource(TychoResource):
    """ System termination. Given a GUID for a Tycho system, use Tycho core to eliminate all
    components comprising the running system."""
    def post(self):
        """
        Delete a system based on a name.
        ---
        tag: start
        description: Delete a system on the compute fabric.
        requestBody:
            description: System start message.
            required: true
            content:
                application/json:
                    schema:
                        $ref: '#/components/schemas/DeleteRequest'
        responses:
            '200':
                description: Success
                content:
                    application/json:
                        schema:
                            type: string
                            example: "Successfully validated"
            '400':
                description: Malformed message
                content:
                    text/plain:
                        schema:
                            type: string

        """
        response = {}
        system_name=None
        try:
            logger.debug (f"delete-request: {json.dumps(request.json, indent=2)}")
            self.validate (request, component="DeleteRequest")
            system_name = request.json['name']
            response = self.create_response (
                result=tycho().get_compute().delete (system_name),
                message=f"Deleted system {system_name}")
        except Exception as e:
            response = self.create_response (
                exception=e,
                message=f"Failed to delete system {system_name}.")
        return response

class StatusSystemResource(TychoResource):
    """ Status executing systems. Given a GUID (or not) determine system status. """
    def post(self):
        """
        Status running systems.
        ---
        tag: start
        description: Status running systems.
        requestBody:
            description: List systems.
            required: true
            content:
                application/json:
                    schema:
                        $ref: '#/components/schemas/StatusRequest'
        responses:
            '200':
                description: Success
                content:
                    application/json:
                        schema:
                            type: string
                            example: "Successfully validated"
            '400':
                description: Malformed message
                content:
                    text/plain:
                        schema:
                            type: string

        """
        response = {}
        try:
            logging.debug (f"list-request: {json.dumps(request.json, indent=2)}")
            self.validate (request, component="StatusRequest") 
            system_name = request.json.get('name', None)
            system_username = request.json.get('username', None)
            response = self.create_response (
                result=tycho().get_compute().status (system_name, system_username),
                message=f"Get status for system {system_name}")
        except Exception as e:
            response = self.create_response (
                exception=e,
                message=f"Failed to get system status.")
        print (json.dumps (response, indent=2))
        return response

""" Register endpoints. """
api.add_resource(StartSystemResource, '/system/start')
api.add_resource(StatusSystemResource, '/system/status')
api.add_resource(DeleteSystemResource, '/system/delete')

if __name__ == "__main__":
   parser = argparse.ArgumentParser(description='Tycho Distributed Compute API')
   parser.add_argument('-b', '--backplane', help='Compute backplane type.', default="kubernetes")
   parser.add_argument('-p', '--port',  type=int, help='Port to run service on.', default=5000)
   parser.add_argument('-d', '--debug', help="Debug log level.", default=False, action='store_true')

   args = parser.parse_args ()

   """ Configure the compute back end. """
   if not Tycho.is_valid_backplane (args.backplane):
       print (f"Unrecognized backplane value: {args.backplane}.")
       print (f"Supported backplanes: {Tycho.supported_backplanes()}")
       parser.print_help ()
       sys.exit (1)
   backplane = args.backplane
   if args.debug:
       debug = True
       logging.basicConfig(level=logging.DEBUG)
   app.run(host='0.0.0.0', port=args.port, threaded=True, debug=args.debug)
